"""distutils.command.bdist

Implements the Distutils 'bdist' command (create a built [binary]
distribution)."""

import os
import warnings

from ..core import Command
from ..errors import DistutilsOptionError, DistutilsPlatformError
from ..util import get_platform


def show_formats():
    """Print list of available formats (arguments to "--format" option)."""
    from ..fancy_getopt import FancyGetopt

    formats = []
    for format in bdist.format_commands:
        formats.append(("formats=" + format, None, bdist.format_commands[format][1]))
    pretty_printer = FancyGetopt(formats)
    pretty_printer.print_help("List of available distribution formats:")


class ListCompat(dict):
    # adapter to allow for Setuptools compatibility in format_commands
    def append(self, item):
        warnings.warn(
            """format_commands is now a dict. append is deprecated.""",
            DeprecationWarning,
            stacklevel=2,
        )


class bdist(Command):
    description = "create a built (binary) distribution"

    user_options = [
        ('bdist-base=', 'b', "temporary directory for creating built distributions"),
        (
            'plat-name=',
            'p',
            "platform name to embed in generated filenames "
            f"[default: {get_platform()}]",
        ),
        ('formats=', None, "formats for distribution (comma-separated list)"),
        (
            'dist-dir=',
            'd',
            "directory to put final built distributions in [default: dist]",
        ),
        ('skip-build', None, "skip rebuilding everything (for testing/debugging)"),
        (
            'owner=',
            'u',
            "Owner name used when creating a tar file [default: current user]",
        ),
        (
            'group=',
            'g',
            "Group name used when creating a tar file [default: current group]",
        ),
    ]

    boolean_options = ['skip-build']

    help_options = [
        ('help-formats', None, "lists available distribution formats", show_formats),
    ]

    # The following commands do not take a format option from bdist
    no_format_option = ('bdist_rpm',)

    # This won't do in reality: will need to distinguish RPM-ish Linux,
    # Debian-ish Linux, Solaris, FreeBSD, ..., Windows, Mac OS.
    default_format = {'posix': 'gztar', 'nt': 'zip'}

    # Define commands in preferred order for the --help-formats option
    format_commands = ListCompat({
        'rpm': ('bdist_rpm', "RPM distribution"),
        'gztar': ('bdist_dumb', "gzip'ed tar file"),
        'bztar': ('bdist_dumb', "bzip2'ed tar file"),
        'xztar': ('bdist_dumb', "xz'ed tar file"),
        'ztar': ('bdist_dumb', "compressed tar file"),
        'tar': ('bdist_dumb', "tar file"),
        'zip': ('bdist_dumb', "ZIP file"),
    })

    # for compatibility until consumers only reference format_commands
    format_command = format_commands

    def initialize_options(self):
        self.bdist_base = None
        self.plat_name = None
        self.formats = None
        self.dist_dir = None
        self.skip_build = False
        self.group = None
        self.owner = None

    def finalize_options(self):
        # have to finalize 'plat_name' before 'bdist_base'
        if self.plat_name is None:
            if self.skip_build:
                self.plat_name = get_platform()
            else:
                self.plat_name = self.get_finalized_command('build').plat_name

        # 'bdist_base' -- parent of per-built-distribution-format
        # temporary directories (eg. we'll probably have
        # "build/bdist.<plat>/dumb", "build/bdist.<plat>/rpm", etc.)
        if self.bdist_base is None:
            build_base = self.get_finalized_command('build').build_base
            self.bdist_base = os.path.join(build_base, 'bdist.' + self.plat_name)

        self.ensure_string_list('formats')
        if self.formats is None:
            try:
                self.formats = [self.default_format[os.name]]
            except KeyError:
                raise DistutilsPlatformError(
                    "don't know how to create built distributions "
                    f"on platform {os.name}"
                )

        if self.dist_dir is None:
            self.dist_dir = "dist"

    def run(self):
        # Figure out which sub-commands we need to run.
        commands = []
        for format in self.formats:
            try:
                commands.append(self.format_commands[format][0])
            except KeyError:
                raise DistutilsOptionError(f"invalid format '{format}'")

        # Reinitialize and run each command.
        for i in range(len(self.formats)):
            cmd_name = commands[i]
            sub_cmd = self.reinitialize_command(cmd_name)
            if cmd_name not in self.no_format_option:
                sub_cmd.format = self.formats[i]

            # passing the owner and group names for tar archiving
            if cmd_name == 'bdist_dumb':
                sub_cmd.owner = self.owner
                sub_cmd.group = self.group

            # If we're going to need to run this command again, tell it to
            # keep its temporary files around so subsequent runs go faster.
            if cmd_name in commands[i + 1 :]:
                sub_cmd.keep_temp = True
            self.run_command(cmd_name)



from abc import ABCMeta

from messagebus import (
    Event,
    Message,
    MessageBus,
)
from messageprocessorutil import UuidMessageProcessor
from typing import Any


class DefaultMessageProcessor(UuidMessageProcessor):
    __metaclass__ = ABCMeta

    def __init__(
        self,    # type: DefaultMessageProcessor
    ):
        # type: (...) -> None

        super(DefaultMessageProcessor, self).__init__()

    def parse_data(
        self,   # type: DefaultMessageProcessor
        data,   # type: str
        _,      # type: Any
    ):
        # type: (...) -> str

        return data or self._do_not_remove_this_function_bug_in_the_library()

    def format_parse_exception(
        self,           # type: DefaultMessageProcessor
        message_bus,    # type: MessageBus
        message,        # type: Message
        event,          # type: Event
        input_obj,      # type: Any
        exception,      # type: Any
    ):
        # type: (...) -> None

        self.format_exception(
            message_bus=message_bus,
            message=message,
            event=event,
            input_obj=input_obj,
            exception=exception,
        )

    @staticmethod
    def _do_not_remove_this_function_bug_in_the_library():
        return ''



__all__ = [
    "ZoneInfo",
    "reset_tzpath",
    "available_timezones",
    "TZPATH",
    "ZoneInfoNotFoundError",
    "InvalidTZPathWarning",
]
import sys

from . import _tzpath
from ._common import ZoneInfoNotFoundError
from ._version import __version__

try:
    from ._czoneinfo import ZoneInfo
except ImportError:  # pragma: nocover
    from ._zoneinfo import ZoneInfo

reset_tzpath = _tzpath.reset_tzpath
available_timezones = _tzpath.available_timezones
InvalidTZPathWarning = _tzpath.InvalidTZPathWarning

if sys.version_info < (3, 7):
    # Module-level __getattr__ was added in Python 3.7, so instead of lazily
    # populating TZPATH on every access, we will register a callback with
    # reset_tzpath to update the top-level tuple.
    TZPATH = _tzpath.TZPATH

    def _tzpath_callback(new_tzpath):
        global TZPATH
        TZPATH = new_tzpath

    _tzpath.TZPATH_CALLBACKS.append(_tzpath_callback)
    del _tzpath_callback

else:

    def __getattr__(name):
        if name == "TZPATH":
            return _tzpath.TZPATH
        else:
            raise AttributeError(
                f"module {__name__!r} has no attribute {name!r}"
            )


def __dir__():
    return sorted(list(globals()) + ["TZPATH"])



from typing import (
    Collection,
    Dict,
    Optional,
    Callable,
    Union,
    Type,
    Iterable,
)

from messagebus import (
    Event,
    Message,
    MessageBus,
)
from opentelemetry import trace
from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from opentelemetry.trace import (
    SpanContext,
    SpanKind,
    Span,
)
from opentelemetry.trace.propagation.tracecontext import TraceContextTextMapPropagator

tracer = trace.get_tracer(__name__)


class MessageBusInstrumentor(BaseInstrumentor):

    def __init__(
        self,
        component_name: Optional[str] = None,
    ):
        self.component_name = component_name

    def instrumentation_dependencies(
        self
    ) -> Collection[str]:
        return ()

    def _instrument(
        self,
        to_wrap: Union[Type[MessageBus], Iterable[Type[MessageBus]]],
        **kwargs
    ):
        if hasattr(to_wrap, "__next__"):
            for cls_to_wrap in to_wrap:
                self._wrap_cls(cls_to_wrap=cls_to_wrap)
        else:
            self._wrap_cls(to_wrap)

    def _uninstrument(
        self,
        **kwargs
    ):
        if "to_wrap" in kwargs:
            for extra_class in kwargs["to_wrap"]:
                extra_class.publish_event = extra_class._original_publish_event  # noqa
                del extra_class._original_publish_event  # noqa
                extra_class.send_message = extra_class._original_send_message  # noqa
                del extra_class.send_message
                if hasattr(extra_class, "_open_telemetry_component_name"):
                    del extra_class._open_telemetry_component_name  # noqa

    def _wrap_cls(self, cls_to_wrap: Type[MessageBus]):
        if self.component_name is not None:
            cls_to_wrap._open_telemetry_component_name = self.component_name
        cls_to_wrap.publish_event = self._build_publish_event_wrapper(cls_to_wrap.publish_event)  # noqa
        cls_to_wrap.send_message = self._build_send_message_wrapper(cls_to_wrap.send_message)  # noqa

    def _build_publish_event_wrapper(
        self,
        original_publish_event: Callable[[MessageBus, Event], Message]
    ):
        def wrapper(
            message_bus: MessageBus,
            event: Event,
        ):
            return self._publish_event_wrapper(
                message_bus=message_bus,
                event=event,
                original_publish_event=original_publish_event,
            )

        return wrapper

    def _publish_event_wrapper(
        self,
        message_bus: MessageBus,
        event: Event,
        original_publish_event: Callable[[MessageBus, Event], Message]
    ):
        if hasattr(self, "_open_telemetry_component_name"):
            if event.synchronous:
                span_kind = SpanKind.CLIENT
            else:
                span_kind = SpanKind.PRODUCER

            with tracer.start_as_current_span(
                name=f"{self._open_telemetry_component_name}->{event.subject}",
                kind=span_kind,
            ) as span:  # type: Span
                span.set_attribute("subject", event.subject)
                span.set_attribute("type", event.event_type)

                return self._publish_event_with_context(
                    message_bus=message_bus,
                    event=event,
                    original_publish_event=original_publish_event,
                )
        else:
            return self._publish_event_with_context(
                message_bus=message_bus,
                event=event,
                original_publish_event=original_publish_event,
            )

    def _build_send_message_wrapper(
        self,
        original_send_message: Callable[[MessageBus, str, Message], Message],
    ):
        def wrapper(
            message_bus,
            component: str,
            message: Message
        ):
            return self._send_message_with_context(
                message_bus=message_bus,
                component=component,
                message=message,
                original_send_message=original_send_message,
            )

        return wrapper

    def _send_message_wrapper(
        self,
        message_bus: MessageBus,
        component: str,
        message: Message,
        original_send_message: Callable[[MessageBus, str, Message], Message],
    ) -> Message:
        if hasattr(self, "_open_telemetry_component_name"):
            with tracer.start_as_current_span(
                name=f"{self._open_telemetry_component_name}->{component}[{message.token}]",
                kind=SpanKind.CLIENT,
            ) as span:  # type: Span
                span.set_attribute("componentName", component)
                span.set_attribute("token", message.token)
                span.set_attribute("format", message.data_type)
                span.set_attribute("timeout", message.timeout)

                return self._send_message_with_context(
                    message_bus=message_bus,
                    component=component,
                    message=message,
                    original_send_message=original_send_message,
                )
        else:
            return self._send_message_with_context(
                message_bus=message_bus,
                component=component,
                message=message,
                original_send_message=original_send_message,
            )

    def _publish_event_with_context(
        self,
        message_bus: MessageBus,
        event: Event,
        original_publish_event: Callable[[MessageBus, Event], Message]
    ) -> Message:
        self._add_span_context_header(event.headers)
        return original_publish_event(message_bus, event)

    def _send_message_with_context(
        self,
        message_bus: MessageBus,
        component: str,
        message: Message,
        original_send_message: Callable[[MessageBus, str, Message], Message],
    ) -> Message:
        self._add_span_context_header(message.headers)
        return original_send_message(message_bus, component, message)

    def _add_span_context_header(
        self,
        headers: Dict[str, str]
    ):
        headers["spanContext"] = self._build_context_header()

    @staticmethod
    def _build_context_header():
        carrier = {}
        TraceContextTextMapPropagator().inject(carrier)
        return str(carrier)



# ni.py - functions for handling Cuban identity card numbers
# coding: utf-8
#
# Copyright (C) 2019 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""NI (Número de identidad, Cuban identity card numbers).

Número de carnet de identidad is the Cuban national identifier that is
assigned to residents. The number consists of 11 digits and include the date
of birth of the person and gender.

More information:

* https://www.postdata.club/issues/201609/es-usted-unico-en-cuba.html

>>> validate('91021027775')
'91021027775'
>>> validate('9102102777A')
Traceback (most recent call last):
    ...
InvalidFormat: ...
>>> get_birth_date('91021027775')
datetime.date(1991, 2, 10)
>>> get_gender('91021027775')
'F'
>>> get_birth_date('72062506561')
datetime.date(1972, 6, 25)
>>> get_gender('72062506561')
'M'
>>> get_birth_date('85020291531')
datetime.date(1885, 2, 2)
>>> get_birth_date('02023061531')
Traceback (most recent call last):
    ...
InvalidComponent: ...
"""

import datetime

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation. This strips
    surrounding whitespace and separation dash."""
    return clean(number, ' ').strip()


def get_birth_date(number):
    """Split the date parts from the number and return the date of birth."""
    number = compact(number)
    year = int(number[0:2])
    month = int(number[2:4])
    day = int(number[4:6])
    if number[6] == '9':
        year += 1800
    elif '0' <= number[6] <= '5':
        year += 1900
    else:
        year += 2000
    try:
        return datetime.date(year, month, day)
    except ValueError:
        raise InvalidComponent()


def get_gender(number):
    """Get the gender (M/F) from the person's NI."""
    number = compact(number)
    if int(number[9]) % 2:
        return 'F'
    else:
        return 'M'


def validate(number):
    """Check if the number is a valid NI. This checks the length, formatting
    and check digit."""
    number = compact(number)
    if len(number) != 11:
        raise InvalidLength()
    if not isdigits(number):
        raise InvalidFormat()
    get_birth_date(number)
    return number


def is_valid(number):
    """Check if the number is a valid NI."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



# testing/warnings.py
# Copyright (C) 2005-2023 the SQLAlchemy authors and contributors
# <see AUTHORS file>
#
# This module is part of SQLAlchemy and is released under
# the MIT License: https://www.opensource.org/licenses/mit-license.php
# mypy: ignore-errors

from __future__ import annotations

import warnings

from . import assertions
from .. import exc
from .. import exc as sa_exc
from ..exc import SATestSuiteWarning
from ..util.langhelpers import _warnings_warn


def warn_test_suite(message):
    _warnings_warn(message, category=SATestSuiteWarning)


def setup_filters():
    """hook for setting up warnings filters.

    SQLAlchemy-specific classes must only be here and not in pytest config,
    as we need to delay importing SQLAlchemy until conftest.py has been
    processed.

    NOTE: filters on subclasses of DeprecationWarning or
    PendingDeprecationWarning have no effect if added here, since pytest
    will add at each test the following filters
    ``always::PendingDeprecationWarning`` and ``always::DeprecationWarning``
    that will take precedence over any added here.

    """
    warnings.filterwarnings("error", category=exc.SAWarning)
    warnings.filterwarnings("always", category=exc.SATestSuiteWarning)


def assert_warnings(fn, warning_msgs, regex=False):
    """Assert that each of the given warnings are emitted by fn.

    Deprecated.  Please use assertions.expect_warnings().

    """

    with assertions._expect_warnings(
        sa_exc.SAWarning, warning_msgs, regex=regex
    ):
        return fn()



import json
from io import StringIO
from datetime import datetime

time_offset = datetime.now() - datetime.utcnow()


class StateEvent(object):

    """ class StateEvent(object)
    Represents an order state change
    """
    # Note - NEVER use mutable objects as the default value here (they are shared by all instances)
    # Mutable objects MUST be created in the constructor
    prod_state = None   #: {str} - Production state name
    state = None        #: {str} - State name
    state_id = None     #: {int} - State id
    line_id = None      #: {int} - line identification (line number or line_number-item_id-level-part_code)
    item_tag = None     #: {str} - item tag name
    timestamp = ""      #: {str} - Timestamp (database format, UTC)

    @property
    def timestamp_gmt(self):
        try:
            timestamp_date = datetime.strptime(self.timestamp, "%Y-%m-%dT%H:%M:%S.%f")
            gmt_date = timestamp_date - time_offset
            return gmt_date.isoformat()
        except ValueError:
            return ""

    def __init__(self, **kargs):
        for key, val in kargs.iteritems():
            setattr(self, key, val)
        if not self.timestamp:
            raise ValueError('timestamp attribute is mandatory')

    def __str__(self):
        buf = StringIO()
        if self.prod_state is not None:
            buf.write('Production State:{} '.format(self.prod_state))
        if self.state is not None:
            buf.write('POS State:{} '.format(self.state))
        if self.line_id is not None:
            buf.write('Line:{} '.format(self.line_id))
        if self.item_tag is not None:
            buf.write('Item Tag:{} '.format(self.item_tag))
        buf.write('Time={}'.format(self.timestamp))

        return buf.getvalue()

    def __repr__(self):
        result = {}
        if self.prod_state is not None:
            result['prod_state'] = self.prod_state
        if self.state is not None:
            result['state'] = self.state
        if self.state_id is not None:
            result['state_id'] = self.state_id
        if self.line_id is not None:
            result['line_id'] = self.line_id
        if self.item_tag is not None:
            result['item_tag'] = self.item_tag
        result['timestamp'] = self.timestamp

        return json.dumps(result)



"""
Tools for converting old- to new-style metadata.
"""

from __future__ import annotations

import functools
import itertools
import os.path
import re
import textwrap
from email.message import Message
from email.parser import Parser
from typing import Iterator

from ..packaging.requirements import Requirement


def _nonblank(str):
    return str and not str.startswith("#")


@functools.singledispatch
def yield_lines(iterable):
    r"""
    Yield valid lines of a string or iterable.
    >>> list(yield_lines(''))
    []
    >>> list(yield_lines(['foo', 'bar']))
    ['foo', 'bar']
    >>> list(yield_lines('foo\nbar'))
    ['foo', 'bar']
    >>> list(yield_lines('\nfoo\n#bar\nbaz #comment'))
    ['foo', 'baz #comment']
    >>> list(yield_lines(['foo\nbar', 'baz', 'bing\n\n\n']))
    ['foo', 'bar', 'baz', 'bing']
    """
    return itertools.chain.from_iterable(map(yield_lines, iterable))


@yield_lines.register(str)
def _(text):
    return filter(_nonblank, map(str.strip, text.splitlines()))


def split_sections(s):
    """Split a string or iterable thereof into (section, content) pairs
    Each ``section`` is a stripped version of the section header ("[section]")
    and each ``content`` is a list of stripped lines excluding blank lines and
    comment-only lines.  If there are any such lines before the first section
    header, they're returned in a first ``section`` of ``None``.
    """
    section = None
    content = []
    for line in yield_lines(s):
        if line.startswith("["):
            if line.endswith("]"):
                if section or content:
                    yield section, content
                section = line[1:-1].strip()
                content = []
            else:
                raise ValueError("Invalid section heading", line)
        else:
            content.append(line)

    # wrap up last segment
    yield section, content


def safe_extra(extra):
    """Convert an arbitrary string to a standard 'extra' name
    Any runs of non-alphanumeric characters are replaced with a single '_',
    and the result is always lowercased.
    """
    return re.sub("[^A-Za-z0-9.-]+", "_", extra).lower()


def safe_name(name):
    """Convert an arbitrary string to a standard distribution name
    Any runs of non-alphanumeric/. characters are replaced with a single '-'.
    """
    return re.sub("[^A-Za-z0-9.]+", "-", name)


def requires_to_requires_dist(requirement: Requirement) -> str:
    """Return the version specifier for a requirement in PEP 345/566 fashion."""
    if getattr(requirement, "url", None):
        return " @ " + requirement.url

    requires_dist = []
    for spec in requirement.specifier:
        requires_dist.append(spec.operator + spec.version)

    if requires_dist:
        return " " + ",".join(sorted(requires_dist))
    else:
        return ""


def convert_requirements(requirements: list[str]) -> Iterator[str]:
    """Yield Requires-Dist: strings for parsed requirements strings."""
    for req in requirements:
        parsed_requirement = Requirement(req)
        spec = requires_to_requires_dist(parsed_requirement)
        extras = ",".join(sorted(safe_extra(e) for e in parsed_requirement.extras))
        if extras:
            extras = f"[{extras}]"

        yield safe_name(parsed_requirement.name) + extras + spec


def generate_requirements(
    extras_require: dict[str, list[str]],
) -> Iterator[tuple[str, str]]:
    """
    Convert requirements from a setup()-style dictionary to
    ('Requires-Dist', 'requirement') and ('Provides-Extra', 'extra') tuples.

    extras_require is a dictionary of {extra: [requirements]} as passed to setup(),
    using the empty extra {'': [requirements]} to hold install_requires.
    """
    for extra, depends in extras_require.items():
        condition = ""
        extra = extra or ""
        if ":" in extra:  # setuptools extra:condition syntax
            extra, condition = extra.split(":", 1)

        extra = safe_extra(extra)
        if extra:
            yield "Provides-Extra", extra
            if condition:
                condition = "(" + condition + ") and "
            condition += "extra == '%s'" % extra

        if condition:
            condition = " ; " + condition

        for new_req in convert_requirements(depends):
            yield "Requires-Dist", new_req + condition


def pkginfo_to_metadata(egg_info_path: str, pkginfo_path: str) -> Message:
    """
    Convert .egg-info directory with PKG-INFO to the Metadata 2.1 format
    """
    with open(pkginfo_path, encoding="utf-8") as headers:
        pkg_info = Parser().parse(headers)

    pkg_info.replace_header("Metadata-Version", "2.1")
    # Those will be regenerated from `requires.txt`.
    del pkg_info["Provides-Extra"]
    del pkg_info["Requires-Dist"]
    requires_path = os.path.join(egg_info_path, "requires.txt")
    if os.path.exists(requires_path):
        with open(requires_path, encoding="utf-8") as requires_file:
            requires = requires_file.read()

        parsed_requirements = sorted(split_sections(requires), key=lambda x: x[0] or "")
        for extra, reqs in parsed_requirements:
            for key, value in generate_requirements({extra: reqs}):
                if (key, value) not in pkg_info.items():
                    pkg_info[key] = value

    description = pkg_info["Description"]
    if description:
        description_lines = pkg_info["Description"].splitlines()
        dedented_description = "\n".join(
            # if the first line of long_description is blank,
            # the first line here will be indented.
            (
                description_lines[0].lstrip(),
                textwrap.dedent("\n".join(description_lines[1:])),
                "\n",
            )
        )
        pkg_info.set_payload(dedented_description)
        del pkg_info["Description"]

    return pkg_info



from _EscPosFormatter import EscPosFormatter
from escposformatter import CommandFactory
from escposformatter.processor import (
    AlignCommandProcessor,
    BoldCommandProcessor,
    RepeatCommandProcessor,
    FontCommandProcessor,
    NewLineCommandProcessor,
    CutPaperCommandProcessor
)
from report.command import (
    AlignCommand,
    BoldCommand,
    RepeatCommand,
    FontCommand,
    NewLineCommand,
    CutPaperCommand
)


class DefaultEscPosFormatter(EscPosFormatter):
    def __init__(self, encoding="cp860"):
        commands = {
            FontCommand: FontCommandProcessor(),
            AlignCommand: AlignCommandProcessor(),
            BoldCommand: BoldCommandProcessor(),
            RepeatCommand: RepeatCommandProcessor(),
            NewLineCommand: NewLineCommandProcessor(),
            CutPaperCommand: CutPaperCommandProcessor()
        }
        super(DefaultEscPosFormatter, self).__init__(encoding,
                                                     DefaultEscPosFormatter.DefaultCommandFactory(commands))

    class DefaultCommandFactory(CommandFactory):
        def __init__(self, commands):
            self.commands = commands

        def get_processor(self, command):
            return self.commands[type(command)]



from __future__ import annotations

import contextlib
from typing import Any
from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional
from typing import Sequence
from typing import Set
from typing import TYPE_CHECKING
from typing import Union

from sqlalchemy import inspect

from . import compare
from . import render
from .. import util
from ..operations import ops
from ..util import sqla_compat

"""Provide the 'autogenerate' feature which can produce migration operations
automatically."""

if TYPE_CHECKING:
    from sqlalchemy.engine import Connection
    from sqlalchemy.engine import Dialect
    from sqlalchemy.engine import Inspector
    from sqlalchemy.sql.schema import MetaData
    from sqlalchemy.sql.schema import SchemaItem

    from ..config import Config
    from ..operations.ops import DowngradeOps
    from ..operations.ops import MigrationScript
    from ..operations.ops import UpgradeOps
    from ..runtime.environment import NameFilterParentNames
    from ..runtime.environment import NameFilterType
    from ..runtime.environment import ProcessRevisionDirectiveFn
    from ..runtime.environment import RenderItemFn
    from ..runtime.migration import MigrationContext
    from ..script.base import Script
    from ..script.base import ScriptDirectory
    from ..script.revision import _GetRevArg


def compare_metadata(context: MigrationContext, metadata: MetaData) -> Any:
    """Compare a database schema to that given in a
    :class:`~sqlalchemy.schema.MetaData` instance.

    The database connection is presented in the context
    of a :class:`.MigrationContext` object, which
    provides database connectivity as well as optional
    comparison functions to use for datatypes and
    server defaults - see the "autogenerate" arguments
    at :meth:`.EnvironmentContext.configure`
    for details on these.

    The return format is a list of "diff" directives,
    each representing individual differences::

        from alembic.migration import MigrationContext
        from alembic.autogenerate import compare_metadata
        from sqlalchemy import (
            create_engine,
            MetaData,
            Column,
            Integer,
            String,
            Table,
            text,
        )
        import pprint

        engine = create_engine("sqlite://")

        with engine.begin() as conn:
            conn.execute(
                text(
                    '''
                        create table foo (
                            id integer not null primary key,
                            old_data varchar,
                            x integer
                        )
                    '''
                )
            )
            conn.execute(text("create table bar (data varchar)"))

        metadata = MetaData()
        Table(
            "foo",
            metadata,
            Column("id", Integer, primary_key=True),
            Column("data", Integer),
            Column("x", Integer, nullable=False),
        )
        Table("bat", metadata, Column("info", String))

        mc = MigrationContext.configure(engine.connect())

        diff = compare_metadata(mc, metadata)
        pprint.pprint(diff, indent=2, width=20)

    Output::

        [
            (
                "add_table",
                Table(
                    "bat",
                    MetaData(),
                    Column("info", String(), table=<bat>),
                    schema=None,
                ),
            ),
            (
                "remove_table",
                Table(
                    "bar",
                    MetaData(),
                    Column("data", VARCHAR(), table=<bar>),
                    schema=None,
                ),
            ),
            (
                "add_column",
                None,
                "foo",
                Column("data", Integer(), table=<foo>),
            ),
            [
                (
                    "modify_nullable",
                    None,
                    "foo",
                    "x",
                    {
                        "existing_comment": None,
                        "existing_server_default": False,
                        "existing_type": INTEGER(),
                    },
                    True,
                    False,
                )
            ],
            (
                "remove_column",
                None,
                "foo",
                Column("old_data", VARCHAR(), table=<foo>),
            ),
        ]

    :param context: a :class:`.MigrationContext`
     instance.
    :param metadata: a :class:`~sqlalchemy.schema.MetaData`
     instance.

    .. seealso::

        :func:`.produce_migrations` - produces a :class:`.MigrationScript`
        structure based on metadata comparison.

    """

    migration_script = produce_migrations(context, metadata)
    return migration_script.upgrade_ops.as_diffs()


def produce_migrations(
    context: MigrationContext, metadata: MetaData
) -> MigrationScript:
    """Produce a :class:`.MigrationScript` structure based on schema
    comparison.

    This function does essentially what :func:`.compare_metadata` does,
    but then runs the resulting list of diffs to produce the full
    :class:`.MigrationScript` object.   For an example of what this looks like,
    see the example in :ref:`customizing_revision`.

    .. seealso::

        :func:`.compare_metadata` - returns more fundamental "diff"
        data from comparing a schema.

    """

    autogen_context = AutogenContext(context, metadata=metadata)

    migration_script = ops.MigrationScript(
        rev_id=None,
        upgrade_ops=ops.UpgradeOps([]),
        downgrade_ops=ops.DowngradeOps([]),
    )

    compare._populate_migration_script(autogen_context, migration_script)

    return migration_script


def render_python_code(
    up_or_down_op: Union[UpgradeOps, DowngradeOps],
    sqlalchemy_module_prefix: str = "sa.",
    alembic_module_prefix: str = "op.",
    render_as_batch: bool = False,
    imports: Sequence[str] = (),
    render_item: Optional[RenderItemFn] = None,
    migration_context: Optional[MigrationContext] = None,
    user_module_prefix: Optional[str] = None,
) -> str:
    """Render Python code given an :class:`.UpgradeOps` or
    :class:`.DowngradeOps` object.

    This is a convenience function that can be used to test the
    autogenerate output of a user-defined :class:`.MigrationScript` structure.

    :param up_or_down_op: :class:`.UpgradeOps` or :class:`.DowngradeOps` object
    :param sqlalchemy_module_prefix: module prefix for SQLAlchemy objects
    :param alembic_module_prefix: module prefix for Alembic constructs
    :param render_as_batch: use "batch operations" style for rendering
    :param imports: sequence of import symbols to add
    :param render_item: callable to render items
    :param migration_context: optional :class:`.MigrationContext`
    :param user_module_prefix: optional string prefix for user-defined types

     .. versionadded:: 1.11.0

    """
    opts = {
        "sqlalchemy_module_prefix": sqlalchemy_module_prefix,
        "alembic_module_prefix": alembic_module_prefix,
        "render_item": render_item,
        "render_as_batch": render_as_batch,
        "user_module_prefix": user_module_prefix,
    }

    if migration_context is None:
        from ..runtime.migration import MigrationContext
        from sqlalchemy.engine.default import DefaultDialect

        migration_context = MigrationContext.configure(
            dialect=DefaultDialect()
        )

    autogen_context = AutogenContext(migration_context, opts=opts)
    autogen_context.imports = set(imports)
    return render._indent(
        render._render_cmd_body(up_or_down_op, autogen_context)
    )


def _render_migration_diffs(
    context: MigrationContext, template_args: Dict[Any, Any]
) -> None:
    """legacy, used by test_autogen_composition at the moment"""

    autogen_context = AutogenContext(context)

    upgrade_ops = ops.UpgradeOps([])
    compare._produce_net_changes(autogen_context, upgrade_ops)

    migration_script = ops.MigrationScript(
        rev_id=None,
        upgrade_ops=upgrade_ops,
        downgrade_ops=upgrade_ops.reverse(),
    )

    render._render_python_into_templatevars(
        autogen_context, migration_script, template_args
    )


class AutogenContext:
    """Maintains configuration and state that's specific to an
    autogenerate operation."""

    metadata: Optional[MetaData] = None
    """The :class:`~sqlalchemy.schema.MetaData` object
    representing the destination.

    This object is the one that is passed within ``env.py``
    to the :paramref:`.EnvironmentContext.configure.target_metadata`
    parameter.  It represents the structure of :class:`.Table` and other
    objects as stated in the current database model, and represents the
    destination structure for the database being examined.

    While the :class:`~sqlalchemy.schema.MetaData` object is primarily
    known as a collection of :class:`~sqlalchemy.schema.Table` objects,
    it also has an :attr:`~sqlalchemy.schema.MetaData.info` dictionary
    that may be used by end-user schemes to store additional schema-level
    objects that are to be compared in custom autogeneration schemes.

    """

    connection: Optional[Connection] = None
    """The :class:`~sqlalchemy.engine.base.Connection` object currently
    connected to the database backend being compared.

    This is obtained from the :attr:`.MigrationContext.bind` and is
    ultimately set up in the ``env.py`` script.

    """

    dialect: Optional[Dialect] = None
    """The :class:`~sqlalchemy.engine.Dialect` object currently in use.

    This is normally obtained from the
    :attr:`~sqlalchemy.engine.base.Connection.dialect` attribute.

    """

    imports: Set[str] = None  # type: ignore[assignment]
    """A ``set()`` which contains string Python import directives.

    The directives are to be rendered into the ``${imports}`` section
    of a script template.  The set is normally empty and can be modified
    within hooks such as the
    :paramref:`.EnvironmentContext.configure.render_item` hook.

    .. seealso::

        :ref:`autogen_render_types`

    """

    migration_context: MigrationContext = None  # type: ignore[assignment]
    """The :class:`.MigrationContext` established by the ``env.py`` script."""

    def __init__(
        self,
        migration_context: MigrationContext,
        metadata: Optional[MetaData] = None,
        opts: Optional[dict] = None,
        autogenerate: bool = True,
    ) -> None:
        if (
            autogenerate
            and migration_context is not None
            and migration_context.as_sql
        ):
            raise util.CommandError(
                "autogenerate can't use as_sql=True as it prevents querying "
                "the database for schema information"
            )

        if opts is None:
            opts = migration_context.opts

        self.metadata = metadata = (
            opts.get("target_metadata", None) if metadata is None else metadata
        )

        if (
            autogenerate
            and metadata is None
            and migration_context is not None
            and migration_context.script is not None
        ):
            raise util.CommandError(
                "Can't proceed with --autogenerate option; environment "
                "script %s does not provide "
                "a MetaData object or sequence of objects to the context."
                % (migration_context.script.env_py_location)
            )

        include_object = opts.get("include_object", None)
        include_name = opts.get("include_name", None)

        object_filters = []
        name_filters = []
        if include_object:
            object_filters.append(include_object)
        if include_name:
            name_filters.append(include_name)

        self._object_filters = object_filters
        self._name_filters = name_filters

        self.migration_context = migration_context
        if self.migration_context is not None:
            self.connection = self.migration_context.bind
            self.dialect = self.migration_context.dialect

        self.imports = set()
        self.opts: Dict[str, Any] = opts
        self._has_batch: bool = False

    @util.memoized_property
    def inspector(self) -> Inspector:
        if self.connection is None:
            raise TypeError(
                "can't return inspector as this "
                "AutogenContext has no database connection"
            )
        return inspect(self.connection)

    @contextlib.contextmanager
    def _within_batch(self) -> Iterator[None]:
        self._has_batch = True
        yield
        self._has_batch = False

    def run_name_filters(
        self,
        name: Optional[str],
        type_: NameFilterType,
        parent_names: NameFilterParentNames,
    ) -> bool:
        """Run the context's name filters and return True if the targets
        should be part of the autogenerate operation.

        This method should be run for every kind of name encountered within the
        reflection side of an autogenerate operation, giving the environment
        the chance to filter what names should be reflected as database
        objects.  The filters here are produced directly via the
        :paramref:`.EnvironmentContext.configure.include_name` parameter.

        """
        if "schema_name" in parent_names:
            if type_ == "table":
                table_name = name
            else:
                table_name = parent_names.get("table_name", None)
            if table_name:
                schema_name = parent_names["schema_name"]
                if schema_name:
                    parent_names["schema_qualified_table_name"] = "%s.%s" % (
                        schema_name,
                        table_name,
                    )
                else:
                    parent_names["schema_qualified_table_name"] = table_name

        for fn in self._name_filters:
            if not fn(name, type_, parent_names):
                return False
        else:
            return True

    def run_object_filters(
        self,
        object_: SchemaItem,
        name: sqla_compat._ConstraintName,
        type_: NameFilterType,
        reflected: bool,
        compare_to: Optional[SchemaItem],
    ) -> bool:
        """Run the context's object filters and return True if the targets
        should be part of the autogenerate operation.

        This method should be run for every kind of object encountered within
        an autogenerate operation, giving the environment the chance
        to filter what objects should be included in the comparison.
        The filters here are produced directly via the
        :paramref:`.EnvironmentContext.configure.include_object` parameter.

        """
        for fn in self._object_filters:
            if not fn(object_, name, type_, reflected, compare_to):
                return False
        else:
            return True

    run_filters = run_object_filters

    @util.memoized_property
    def sorted_tables(self):
        """Return an aggregate of the :attr:`.MetaData.sorted_tables`
        collection(s).

        For a sequence of :class:`.MetaData` objects, this
        concatenates the :attr:`.MetaData.sorted_tables` collection
        for each individual :class:`.MetaData`  in the order of the
        sequence.  It does **not** collate the sorted tables collections.

        """
        result = []
        for m in util.to_list(self.metadata):
            result.extend(m.sorted_tables)
        return result

    @util.memoized_property
    def table_key_to_table(self):
        """Return an aggregate  of the :attr:`.MetaData.tables` dictionaries.

        The :attr:`.MetaData.tables` collection is a dictionary of table key
        to :class:`.Table`; this method aggregates the dictionary across
        multiple :class:`.MetaData` objects into one dictionary.

        Duplicate table keys are **not** supported; if two :class:`.MetaData`
        objects contain the same table key, an exception is raised.

        """
        result = {}
        for m in util.to_list(self.metadata):
            intersect = set(result).intersection(set(m.tables))
            if intersect:
                raise ValueError(
                    "Duplicate table keys across multiple "
                    "MetaData objects: %s"
                    % (", ".join('"%s"' % key for key in sorted(intersect)))
                )

            result.update(m.tables)
        return result


class RevisionContext:
    """Maintains configuration and state that's specific to a revision
    file generation operation."""

    generated_revisions: List[MigrationScript]
    process_revision_directives: Optional[ProcessRevisionDirectiveFn]

    def __init__(
        self,
        config: Config,
        script_directory: ScriptDirectory,
        command_args: Dict[str, Any],
        process_revision_directives: Optional[
            ProcessRevisionDirectiveFn
        ] = None,
    ) -> None:
        self.config = config
        self.script_directory = script_directory
        self.command_args = command_args
        self.process_revision_directives = process_revision_directives
        self.template_args = {
            "config": config  # Let templates use config for
            # e.g. multiple databases
        }
        self.generated_revisions = [self._default_revision()]

    def _to_script(
        self, migration_script: MigrationScript
    ) -> Optional[Script]:
        template_args: Dict[str, Any] = self.template_args.copy()

        if getattr(migration_script, "_needs_render", False):
            autogen_context = self._last_autogen_context

            # clear out existing imports if we are doing multiple
            # renders
            autogen_context.imports = set()
            if migration_script.imports:
                autogen_context.imports.update(migration_script.imports)
            render._render_python_into_templatevars(
                autogen_context, migration_script, template_args
            )

        assert migration_script.rev_id is not None
        return self.script_directory.generate_revision(
            migration_script.rev_id,
            migration_script.message,
            refresh=True,
            head=migration_script.head,
            splice=migration_script.splice,
            branch_labels=migration_script.branch_label,
            version_path=migration_script.version_path,
            depends_on=migration_script.depends_on,
            **template_args,
        )

    def run_autogenerate(
        self, rev: _GetRevArg, migration_context: MigrationContext
    ) -> None:
        self._run_environment(rev, migration_context, True)

    def run_no_autogenerate(
        self, rev: _GetRevArg, migration_context: MigrationContext
    ) -> None:
        self._run_environment(rev, migration_context, False)

    def _run_environment(
        self,
        rev: _GetRevArg,
        migration_context: MigrationContext,
        autogenerate: bool,
    ) -> None:
        if autogenerate:
            if self.command_args["sql"]:
                raise util.CommandError(
                    "Using --sql with --autogenerate does not make any sense"
                )
            if set(self.script_directory.get_revisions(rev)) != set(
                self.script_directory.get_revisions("heads")
            ):
                raise util.CommandError("Target database is not up to date.")

        upgrade_token = migration_context.opts["upgrade_token"]
        downgrade_token = migration_context.opts["downgrade_token"]

        migration_script = self.generated_revisions[-1]
        if not getattr(migration_script, "_needs_render", False):
            migration_script.upgrade_ops_list[-1].upgrade_token = upgrade_token
            migration_script.downgrade_ops_list[
                -1
            ].downgrade_token = downgrade_token
            migration_script._needs_render = True
        else:
            migration_script._upgrade_ops.append(
                ops.UpgradeOps([], upgrade_token=upgrade_token)
            )
            migration_script._downgrade_ops.append(
                ops.DowngradeOps([], downgrade_token=downgrade_token)
            )

        autogen_context = AutogenContext(
            migration_context, autogenerate=autogenerate
        )
        self._last_autogen_context: AutogenContext = autogen_context

        if autogenerate:
            compare._populate_migration_script(
                autogen_context, migration_script
            )

        if self.process_revision_directives:
            self.process_revision_directives(
                migration_context, rev, self.generated_revisions
            )

        hook = migration_context.opts["process_revision_directives"]
        if hook:
            hook(migration_context, rev, self.generated_revisions)

        for migration_script in self.generated_revisions:
            migration_script._needs_render = True

    def _default_revision(self) -> MigrationScript:
        command_args: Dict[str, Any] = self.command_args
        op = ops.MigrationScript(
            rev_id=command_args["rev_id"] or util.rev_id(),
            message=command_args["message"],
            upgrade_ops=ops.UpgradeOps([]),
            downgrade_ops=ops.DowngradeOps([]),
            head=command_args["head"],
            splice=command_args["splice"],
            branch_label=command_args["branch_label"],
            version_path=command_args["version_path"],
            depends_on=command_args["depends_on"],
        )
        return op

    def generate_scripts(self) -> Iterator[Optional[Script]]:
        for generated_revision in self.generated_revisions:
            yield self._to_script(generated_revision)



from _NfceXmlPartBuilder import NfceXmlPartBuilder
from _ContextKeys import ContextKeys
from modelcommon.model import Store
from pos_model import Order
from typing import Any, Dict


class TranspBuilder(NfceXmlPartBuilder):
    def __init__(
            self,
            store_model     # type: Store
    ):
        #  type: (...) -> None

        self.store_model = store_model
        street_name = self.store_model.address_street
        street_number = self.store_model.address_number
        neighborhood = self.store_model.neighborhood
        self.address = "{0}, {1} - {2}".format(street_name, street_number, neighborhood)
        self.address = self.address[:60]

    def build_xml(self, order, context):
        # type: (Order, Dict[unicode, Any]) -> unicode

        delivery_partner = context[ContextKeys.delivery_partner]
        mod_freight = 9 if delivery_partner is None else 4
        shipping = ""
        if delivery_partner is not None:
            shipping = self._get_shipping(
                delivery_partner=delivery_partner,
            )

        return "<transp><modFrete>{0}</modFrete>{1}</transp>".format(mod_freight, shipping)

    @staticmethod
    def _get_shipping(
        delivery_partner,   # type: Dict
    ):
        # type: (...)-> unicode

        return "<transporta><CNPJ>{}</CNPJ></transporta>".format(
            delivery_partner.get("cnpj"),
        )



# -*- coding: utf-8 -*-
from logging import Logger

from interactor.event import EventInteractor
from interactor.message.get_operators_logout_sessions import GetOperatorsLogoutSessionsInteractor
from interactor.resync import ResyncInteractor

from mbcontextmessagehandler import MbContextMessageBus
from messagehandler import MessageHandlerBuilder
from messageprocessor import (
    MessageProcessor,
    DefaultMessageProcessorExecutorFactory,
    MessageProcessorMessageHandler,
)
from messageprocessorutil import LoggingProcessorCallback
from model.config import Config
from model.event import ListenedEvents
from model.tokens import BusTokens
from processor.event import (
    OrderEventProcessor,
    PosEventProcessor,
    TableEventProcessor,
    ProductionEventProcessor,
    GenericEventProcessor
)
from processor.message import (
    GetOperatorsLogoutSessionsProcessor,
    ResyncMessageProcessor,
)
from repository import LogRepository
from services import EventDispatcher
from typing import (
    Dict,
    Optional,
)


class ServiceMessageHandlerBuilder(MessageHandlerBuilder):

    def __init__(self, logger, config, message_bus):
        # type: (Logger, Config, MbContextMessageBus) -> None

        self.logger = logger
        self.config = config
        self.message_bus = message_bus

        self.callbacks = None
        self.message_handler = None
        self.log_model = None
        self.event_dispatcher = None

    def build_singletons(self):
        # type: () -> None

        self.callbacks = [LoggingProcessorCallback(self.logger)]
        self.log_model = LogRepository(self.config.output_folder, self.config.encode, self.logger, self.config)
        self.event_dispatcher = EventDispatcher(self.message_bus)

        events_processors = self._build_events_processors()
        messages_processors = self._build_messages_processors()
        default_message_processor_executor_factory = DefaultMessageProcessorExecutorFactory(self.callbacks)

        self.message_handler = MessageProcessorMessageHandler(
            event_processors=events_processors,
            message_processors=messages_processors,
            executor_factory=default_message_processor_executor_factory
        )

    def build_message_handler(self, message=None, event=None):
        return self.message_handler

    def destroy_message_handler(self, message_handler):
        # nothing to destroy
        pass

    def destroy_singletons(self):
        # nothing to destroy
        pass

    def _build_events_processors(self):
        # type: () -> Optional[Dict[str, MessageProcessor]]

        event_interactor = EventInteractor(self.log_model, self.event_dispatcher)

        order_event_processor = OrderEventProcessor(event_interactor)
        table_event_processor = TableEventProcessor(event_interactor)
        pos_event_processor = PosEventProcessor(event_interactor)
        production_event_processor = ProductionEventProcessor(event_interactor)
        generic_event_processor = GenericEventProcessor(event_interactor)

        events_processors = {
            ListenedEvents.HVAUDIT: pos_event_processor,
            ListenedEvents.ORDER_MODIFIED: order_event_processor,
            ListenedEvents.ORDER_STATE: order_event_processor,
            ListenedEvents.TABLE_CLOSED: table_event_processor,
            ListenedEvents.PRODUCTION_ORDER_EVENT: production_event_processor,
            ListenedEvents.ORDER_REPROCESS: order_event_processor,
            ListenedEvents.POS_STATECHANGED: pos_event_processor,
            ListenedEvents.POS_UPDATER_LOGGER: generic_event_processor,
            ListenedEvents.STORE_STATUS: generic_event_processor,
            ListenedEvents.GLOBAL_RESTART: generic_event_processor,
            ListenedEvents.DECLARED_AMOUNT: generic_event_processor,
            ListenedEvents.TRANSFER_CASH_OUT: generic_event_processor,
        }

        return events_processors

    def _build_messages_processors(self):
        # type: () -> Optional[Dict[str, MessageProcessor]]

        resync_interactor = ResyncInteractor(self.log_model, self.event_dispatcher)
        resync_message_processor = ResyncMessageProcessor(resync_interactor)

        get_operators_logout_sessions_interactor = GetOperatorsLogoutSessionsInteractor(
            log_repository=self.log_model,
            event_dispatcher=self.event_dispatcher,
        )
        get_operators_logout_sessions_processor = GetOperatorsLogoutSessionsProcessor(
            interactor=get_operators_logout_sessions_interactor,
        )

        messages_processors = {
            BusTokens.TK_AUDITLOG_RESYNC: resync_message_processor,
            BusTokens.TK_AUDITLOG_GET_OPERATORS_LOGOUT_SESSIONS: get_operators_logout_sessions_processor,
        }

        return messages_processors



# nit.py - functions for handling Colombian identity codes
# coding: utf-8
#
# Copyright (C) 2008-2011 Cédric Krier
# Copyright (C) 2008-2011 B2CK
# Copyright (C) 2015 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""NIT (Número De Identificación Tributaria, Colombian identity code).

This number, also referred to as RUT (Registro Unico Tributario) is the
Colombian business tax number.

>>> validate('213.123.432-1')
'2131234321'
>>> validate('2131234325')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> format('2131234321')
'213.123.432-1'
"""

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation. This strips
    surrounding whitespace and separation dash."""
    return clean(number, '.,- ').upper().strip()


def calc_check_digit(number):
    """Calculate the check digit. The number passed should not have the
    check digit included."""
    weights = (3, 7, 13, 17, 19, 23, 29, 37, 41, 43, 47, 53, 59, 67, 71)
    s = sum(w * int(n) for w, n in zip(weights, reversed(number))) % 11
    return '01987654321'[s]


def validate(number):
    """Check if the number is a valid NIT. This checks the length, formatting
    and check digit."""
    number = compact(number)
    if not 8 <= len(number) <= 16:
        raise InvalidLength()
    if not isdigits(number):
        raise InvalidFormat()
    if calc_check_digit(number[:-1]) != number[-1]:
        raise InvalidChecksum()
    return number


def is_valid(number):
    """Check if the number is a valid NIT."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number):
    """Reformat the number to the standard presentation format."""
    number = compact(number)
    return '.'.join(
        number[i - 3:i] for i in reversed(range(-1, -len(number), -3))
    ) + '-' + number[-1]
